/**
 * @file artis/kernel/ModelFactory.hpp
 * @author See the AUTHORS file
 */

/*
 * Copyright (C) 2012-2017 ULCO http://www.univ-littoral.fr
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __ARTIS_KERNEL_MODEL_FACTORY_HPP
#define __ARTIS_KERNEL_MODEL_FACTORY_HPP

#include <algorithm>
#include <memory>
#include <mutex>
#include <typeinfo>

#include <boost/core/demangle.hpp>

namespace artis { namespace kernel {

template < typename M >
class ObjectCreator
{
public:
    ObjectCreator()
    { }
    virtual ~ObjectCreator()
    { }

    virtual std::string get() const = 0;

    virtual M* operator()() const = 0;
};

template < typename M, typename I, typename O >
class ModelFactory
{
    typedef ModelFactory < M, I, O > type;

public:
    virtual ~ModelFactory()
    { }

    bool add(const I& id, O* creator)
    {
        std::lock_guard < std::mutex > lock(_mutex);
        return creators.insert(typename Creators::value_type(
                                   creator->get(),
                                   std::make_pair(id, creator))).second;
    }

    M* create(const std::string& id)
    {
        std::lock_guard < std::mutex > lock(_mutex);
        typename Creators::const_iterator it = creators.find(id);

        if (it != creators.end()) {
            return (*it->second.second)();
        } else {
            return nullptr;
        }
    }

    static ModelFactory& factory()
    {
        std::call_once(_flag, [] ()
                       { _instance.reset(new ModelFactory()); });
        return *_instance;
    }

    int make_id()
    {
        std::lock_guard < std::mutex > lock(_mutex);
        return ++id;
    }

    bool remove(const I& id)
    {
/*        std::lock_guard < std::mutex > lock(_mutex);
        typename Creators::const_iterator it = creators.find(id);

        if (it != creators.end()) {
            creators.erase(it);
            return true;
        } else {
            return false;
            } */
        return false;
    }

private:
    ModelFactory()
    { id = -1; }

    typedef std::pair < I, O* > Creator;
    typedef std::map < std::string, Creator > Creators;

    static std::shared_ptr < type > _instance;
    static std::once_flag           _flag;
    std::mutex                      _mutex;

    Creators creators;
    int      id;
};

} }

template < typename M, typename I, typename O >
std::shared_ptr < artis::kernel::ModelFactory < M, I, O > >
artis::kernel::ModelFactory < M, I, O >::_instance;

template < typename M, typename I, typename O >
std::once_flag artis::kernel::ModelFactory < M, I, O >::_flag;

#define DECLARE_MODEL(mdl, type, fct)                                   \
    class creator_##mdl : public artis::kernel::ObjectCreator < type > { \
    public:                                                             \
    creator_##mdl()                                                     \
        {                                                               \
            fct::factory().add(fct::factory().make_id(), this);         \
        }                                                               \
    std::string get() const                                             \
        { return boost::core::demangle(typeid(mdl).name()); }           \
    type* operator()() const                                            \
        { return new mdl(); }                                           \
    static creator_##mdl an_##mdl;                                      \
    };                                                                  \
    creator_##mdl an_##mdl;

#endif
